/*=========================================================================
| (c) 2017  Total Phase, Inc.
|--------------------------------------------------------------------------
| Project : PD Analyzer Sample Code
| File    : capture_usbpd.cs
|--------------------------------------------------------------------------
| Simple Capture Example
|--------------------------------------------------------------------------
| Redistribution and use of this file in source and binary forms, with
| or without modification, are permitted.
|
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
| "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
| LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
| FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
| COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
| INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
| BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
| LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
| ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
| POSSIBILITY OF SUCH DAMAGE.
 ========================================================================*/
using System;
using System.IO;
using TotalPhase;


/*=========================================================================
| CLASS
 ========================================================================*/
public class Capture_usbpd {

    /*=====================================================================
    | CONSTANTS
     ====================================================================*/


    /*=====================================================================
    | STATIC PRINT FUNCTIONS
     ====================================================================*/
    static bool print_pd_info (int idx, ref PdApi.PdReadInfo info,
                               String info_type)
    {
        int cc, pos_err;

        Console.Write("{0:d}, {1:d}, {2:s}, ", idx, info.timestamp, info_type);

        if (info_type != "PD")
            return true;

        cc = (int)((info.events & PdApi.PD_EVENT_USBPD_CC_MASK)
                   >> (int)PdApi.PD_EVENT_USBPD_CC_SHIFT);
        Console.Write("{0:d}, ", (cc + 1));

        if ((info.events & PdApi.PD_EVENT_USBPD_POL_CHANGE) != 0) {
            Console.Write("POL_CHANGE\n");
            return false;
        }
        else if ((info.events & PdApi.PD_EVENT_USBPD_SOP_MASK) ==
                 PdApi.PD_EVENT_USBPD_HARD_RESET) {
            Console.Write("HARD_RESET\n");
            return false;
        }
        else if ((info.events & PdApi.PD_EVENT_USBPD_SOP_MASK) ==
                 PdApi.PD_EVENT_USBPD_CABLE_RESET) {
            Console.Write("CABLE_RESET\n");
            return false;
        }

        pos_err = (int)(info.status & PdApi.PD_STATUS_USBPD_ERR_MASK);
        if (pos_err == PdApi.PD_STATUS_USBPD_ERR_PREAMBLE)
            Console.Write("ERR_PREAMBLE, ");
        else if (pos_err == PdApi.PD_STATUS_USBPD_ERR_SOP)
            Console.Write("ERR_SOP, ");
        else if (pos_err == PdApi.PD_STATUS_USBPD_ERR_HEADER)
            Console.Write("ERR_HEADER, ");
        else if (pos_err == PdApi.PD_STATUS_USBPD_ERR_DATA)
            Console.Write("ERR_DATA, ");
        else if (pos_err == PdApi.PD_STATUS_USBPD_ERR_CRC)
            Console.Write("ERR_CRC, ");
        else if (pos_err == PdApi.PD_STATUS_USBPD_ERR_EOP)
            Console.Write("ERR_EOP, ");

        if ((info.status & PdApi.PD_STATUS_USBPD_ERR_BAD_CRC) != 0)
            Console.Write("BAD_CRC, ");
        else if (pos_err == 0)
            Console.Write("OK, ");

        return true;
    }

    static void print_bits (int idx, ref PdApi.PdReadInfo info, byte[] bits,
                            uint pl, uint bl)
    {
        String[] kcode_str = new String[] {
            "Error", "Error", "Error", "Error", "Error", "Error", "SYNC3", "RST1",
            "Error", "1", "4", "5", "Error", "EOP", "6", "7",
            "Error", "SYNC2", "8", "9", "2", "3", "A", "B",
            "SYNC1", "RST2", "C", "D", "E", "F", "0", "Error"
        };
        int ii, jj, kcode, off, bit;

        if (!print_pd_info(idx, ref info, "PD"))
            return;

        Console.Write("{0:d}, ", pl);

        for (ii = (int)pl; ii < (int)bl; ii += 5) {
            if (ii + 5 > bl)
                break;

            kcode = 0;
            for (jj = 0; jj < 5; ++jj) {
                off = ii + jj;
                bit = (bits[off / 8] >> (7 - (off % 8))) & 0x1;

                kcode |= (bit << jj);
            }

            Console.Write("{0:s} ", kcode_str[kcode]);
        }

        Console.Write("\n");
    }

    static void print_pd_packets (int idx, ref PdApi.PdReadInfo info,
                                  uint preamble, uint header, uint crc,
                                  int len, byte[] data)
    {
        String[] sop_str = new String[] {
            "SOP", "SOP\'", "SOP\"", "SOP\'D", "SOP\"D", "", "", ""
        };
        int pos_err, sop, offset;

        if (!print_pd_info(idx, ref info, "PD"))
            return;

        pos_err = (int)(info.status & PdApi.PD_STATUS_USBPD_ERR_MASK);

        if (pos_err == PdApi.PD_STATUS_USBPD_ERR_PREAMBLE ||
            pos_err == PdApi.PD_STATUS_USBPD_ERR_SOP) {
            Console.Write("\n");
            return;
        }

        sop = (int)(info.events & PdApi.PD_EVENT_USBPD_SOP_MASK);

        Console.Write("{0:s}, ", sop_str[sop]);

        if (pos_err == PdApi.PD_STATUS_USBPD_ERR_HEADER) {
            Console.Write("\n");
            return;
        }

        if ((info.events & PdApi.PD_EVENT_USBPD_EXTENDED_HEADER) != 0) {
            Console.Write("{0:x04};{0:x04}, ",
                   header & PdApi.PD_HEADER_USBPD_STANDARD_MASK,
                   (header & PdApi.PD_HEADER_USBPD_EXTENDED_MASK) >> 16);
        }
        else {
            Console.Write("{0:x04}, ", header & PdApi.PD_HEADER_USBPD_STANDARD_MASK);
        }

        if (len > 0) {
            for (offset = 0; offset < len; ++offset) {
                Console.Write("{0:x02} ", data[offset]);
            }
        }

        Console.Write(", ");

        if (pos_err == PdApi.PD_STATUS_USBPD_ERR_DATA ||
            pos_err == PdApi.PD_STATUS_USBPD_ERR_CRC) {
            Console.Write("\n");
            return;
        }

        if ((info.status & PdApi.PD_STATUS_USBPD_ERR_BAD_CRC) != 0)
            Console.Write("{0:x04}(NOT OK)\n", crc);
        else
            Console.Write("{0:x04}(OK)\n", crc);
    }

    static void print_pd_iv (int idx, ref PdApi.PdReadInfo info, int val)
    {
        String[] iv_type_str = new String[] {
            "Vbus Voltage", "Vbus Current", "Vconn Voltage", "Vconn Current",
            "CC1 Voltage", "CC1 Current", "CC2 Voltage", "CC2 Current",
        };
        int iv_type;

        if (!print_pd_info(idx, ref info, "PD_IV"))
            return;

        iv_type = (int)((info.events & PdApi.PD_EVENT_USBPD_IV_SOURCE_MASK)
                        >> (int)PdApi.PD_EVENT_USBPD_IV_SHIFT);
        Console.Write("{0:s}, {1:d}{2:s}\n",
                      iv_type_str[iv_type], val,
                      ((iv_type & 1) != 0)? "mA": "mV");
    }



    /*=====================================================================
    | STATIC DUMP FUNCTIONS
     ====================================================================*/
    static void dump_pd_raw (int pd, int num_packets)
    {
        int packetnum = 0;
        PdApi.PdReadInfo info = new PdApi.PdReadInfo();
        int ret;
        uint bl = 0, pl = 0;
        byte[] bits = new byte[512];

        Console.Write("index,time(us),PD,status,cc,preamble,bits\n");

        while (num_packets <= 0 || packetnum < num_packets) {
            // maximum bit length of USB PD v3.0 is 2769 bits = 347 bytes
            ret = PdApi.pd_usbpd_read_bits(pd, ref info, ref bl, ref pl,
                                           (uint)bits.Length, bits);
            if (ret == (int)PdStatus.PD_READ_EMPTY)
                continue;

            if (ret < 0) {
                Console.Write("Failed to capture PD on port) {0:d}({1:s})\n",
                              ret, PdApi.pd_status_string(ret));
                break;
            }

            print_bits(packetnum, ref info, bits, pl, bl);

            packetnum += 1;
        }
    }

    static void dump_pd_data (int pd, int num_packets)
    {
        int packetnum = 0;
        PdApi.PdReadInfo info = new PdApi.PdReadInfo();
        int ret;
        uint pl = 0, header = 0, crc = 0;
        byte[] data = new byte[512];

        Console.Write("index,time(us),PD,cc,status,sop,header,data0 ... dataN,crc\n");

        while (num_packets <= 0 || packetnum < num_packets) {
            // maximum bit length of USB PD v3.0 is 2769 bits = 347 bytes
            ret = PdApi.pd_usbpd_read_data(pd, ref info, ref pl,
                                           ref header, ref crc,
                                           (uint)data.Length, data);
            if (ret == (int)PdStatus.PD_READ_EMPTY)
                continue;

            if (ret < 0) {
                Console.Write("Failed to capture PD on port) {0:d}({1:s})\n",
                              ret, PdApi.pd_status_string(ret));
                break;
            }

            print_pd_packets(packetnum, ref info, pl, header, crc, ret, data);

            packetnum += 1;
        }
    }

    static void dump_pd_iv (int pd, int num_packets)
    {
        int packetnum = 0;
        PdApi.PdReadInfo info = new PdApi.PdReadInfo();
        int ret, val = 0;

        Console.Write("index,time(us),PD_IV,type,value\n");

        while (num_packets <= 0 || packetnum < num_packets) {
            ret = PdApi.pd_usbpd_read_iv(pd, ref info, ref val);
            if (ret == (int)PdStatus.PD_READ_EMPTY)
                continue;

            if (ret < 0) {
                Console.Write("Failed to capture PD on port) {0:d}({1:s})\n",
                              ret, PdApi.pd_status_string(ret));
                break;
            }

            print_pd_iv(packetnum, ref info, val);

            packetnum += 1;
        }
    }

    static void dump_pd_all (int pd, int num_packets)
    {
        int packetnum = 0;
        PdApi.PdReadInfo info = new PdApi.PdReadInfo();
        int ret, val = 0;
        uint pl = 0, header = 0, crc = 0;
        byte[] data = new byte[512];

        Console.Write("index,time(us),PD,cc,status,sop,header,data0 ... dataN,crc\n" +
                      "or\n" +
                      "index,time(us),PD_IV,type,value\n");

        while (num_packets <= 0 || packetnum < num_packets) {
            // maximum bit length of USB PD v3.0 is 2769 bits = 347 bytes
            ret = PdApi.pd_usbpd_read_data(pd, ref info, ref pl,
                                           ref header, ref crc,
                                           (uint)data.Length, data);
            if (ret != (int)PdStatus.PD_READ_EMPTY) {
                if (ret < 0) {
                    Console.Write("Failed to capture PD on port) {0:d}({1:s})\n",
                                  ret, PdApi.pd_status_string(ret));
                    break;
                }

                print_pd_packets(packetnum, ref info, pl, header, crc, ret, data);

                packetnum += 1;
            }

            ret = PdApi.pd_usbpd_read_iv(pd, ref info, ref val);
            if (ret != (int)PdStatus.PD_READ_EMPTY) {
                if (ret < 0) {
                    Console.Write("Failed to capture PD on port) {0:d}({1:s})\n",
                                  ret, PdApi.pd_status_string(ret));
                    break;
                }

                print_pd_iv(packetnum, ref info, val);

                packetnum += 1;
            }
        }
    }


    /*=====================================================================
    | USAGE INFORMATION
     ====================================================================*/
    static void print_usage ()
    {
        Console.Write(
"Usage) capture_usbpd port mode num_events\n" +
"Example utility for capturing USB PD data from Power Delivery analyzers.\n" +
"\n" +
"  The parameter mode can be one of 'raw', 'data', 'iv', and 'all'.\n" +
"    'raw' mode shows raw data after bmc decoding, 'data' mode shows pd packets,\n" +
"    'iv' mode shows voltage and current, and 'all' is combined of 'data' and 'iv'\n" +
"\n" +
"  The parameter num_events is set to the number of events to process\n" +
"    before exiting.  If num_events is set to zero, the capture will continue\n" +
"    indefinitely.\n" +
"\n" +
"For product documentation and specifications, see www.totalphase.com.\n");
        Console.Out.Flush();
    }

    /*=====================================================================
    | MAIN PROGRAM
     ====================================================================*/
    public static void Main (String[] args)
    {
        int    pd       = 0;
        int    port     = 0;      // open port 0 by default
        String mode     = "";
        int    num      = 0;

        if (args.Length < 3)
        {
            print_usage();
            Environment.Exit(1);
        }
        port = Convert.ToInt32(args[0]);
        mode = args[1];
        num  = Convert.ToInt32(args[2]);

        // Open the device
        pd = PdApi.pd_open(port);
        if (pd <= 0) {
            Console.Write("Unable to open PD Analyzer on port {0:d}\n", port);
            Console.Write("Error code = {0:d} ({1:s})\n",
                          pd, PdApi.pd_status_string(pd));
            Environment.Exit(1);
        }
        Console.Write("Opened PD Analyzer on port {0:d}\n", port);

        // Start capture
        PdApi.pd_capture_start(pd);

        Console.Write("Start capturing\n");
        if (mode == "raw")
            dump_pd_raw(pd, num);
        if (mode == "data")
            dump_pd_data(pd, num);
        if (mode == "iv")
            dump_pd_iv(pd, num);
        if (mode == "all")
            dump_pd_all(pd, num);

        // Stop capture
        PdApi.pd_capture_stop(pd);

        // Close the device
        PdApi.pd_close(pd);

        return;
    }
}
